/*
 * alloc.c
 *
 * Copyright (C) 2008-2019 Aerospike, Inc.
 *
 * Portions may be licensed to Aerospike, Inc. under one or more contributor
 * license agreements.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/
 */

// Make sure that stdlib.h gives us aligned_alloc().
#define _ISOC11_SOURCE

#include "enhanced_alloc.h"

#include <errno.h>
#include <inttypes.h>
#include <malloc.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <jemalloc/jemalloc.h>

#include <sys/syscall.h>
#include <sys/types.h>

#include "cf_thread.h"
#include "fault.h"

#include "aerospike/as_random.h"
#include "aerospike/ck/ck_pr.h"
#include "citrusleaf/cf_atomic.h"
#include "citrusleaf/cf_clock.h"

#include "warnings.h"

#undef strdup
#undef strndup

#define N_ARENAS 149 // used to be 150; now arena 149 is startup arena
#define PAGE_SZ 4096

#define MAX_SITES 4096
#define MAX_THREADS 1024

#define MULT 3486784401u
#define MULT_INV 3396732273u

#define MULT_64 12157665459056928801ul
#define MULT_INV_64 12381265223964269537ul

#define STR_(x) #x
#define STR(x) STR_(x)

#define MAX_INDENT (32 * 8)

typedef struct site_info_s {
	uint32_t site_id;
	pid_t thread_id;
	size_t size_lo;
	size_t size_hi;
} site_info;

// Old glibc versions don't provide this; work around compiler warning.
void *aligned_alloc(size_t align, size_t sz);

// When fortification is disabled, glibc's headers don't provide this.
int32_t __asprintf_chk(char **res, int32_t flags, const char *form, ...);

const char *jem_malloc_conf = "narenas:" STR(N_ARENAS);

extern size_t je_chunksize_mask;
extern void *je_huge_aalloc(const void *p);

__thread int32_t g_ns_arena = -1;
static __thread int32_t g_ns_tcache = -1;

static const void *g_site_ras[MAX_SITES];
static uint32_t g_n_site_ras;

static site_info g_site_infos[MAX_SITES * MAX_THREADS];
// Start at 1, then we can use site ID 0 to mean "no site ID".
static uint64_t g_n_site_infos = 1;

static __thread uint32_t g_thread_site_infos[MAX_SITES];

bool g_alloc_started = false;
static int32_t g_startup_arena = -1;

static cf_alloc_debug g_debug;
static bool g_indent;

static __thread as_random g_rand = { .initialized = false };

// All the hook_*() functions are invoked from hook functions that hook into
// malloc() and friends for memory accounting purposes.
//
// This means that we have no idea who called us and, for example, which locks
// they hold. Let's be careful when calling back into asd code.

static int32_t
hook_get_arena(const void *p_indent)
{
	// Disregard indent by rounding down to page boundary. Works universally:
	//
	//   - Small / large: chunk's base aligned to 2 MiB && p >= base + 0x1000.
	//   - Huge: p aligned to 2 MiB && MAX_INDENT < 0x1000.
	//
	// A huge allocations is thus rounded to its actual p (aligned to 2 MiB),
	// but a small or large allocation is never rounded to the chunk's base.

	const void *p = (const void *)((uint64_t)p_indent & ~0xffful);

	int32_t **base = (int32_t **)((uint64_t)p & ~je_chunksize_mask);
	int32_t *arena;

	if (base != p) {
		// Small or large allocation.
		arena = base[0];
	}
	else {
		// Huge allocation.
		arena = je_huge_aalloc(p);
	}

	return arena[0];
}

// Map a 64-bit address to a 12-bit site ID.

static uint32_t
hook_get_site_id(const void *ra)
{
	uint32_t site_id = (uint32_t)(uint64_t)ra & (MAX_SITES - 1);

	for (uint32_t i = 0; i < MAX_SITES; ++i) {
		const void *site_ra = ck_pr_load_ptr(g_site_ras + site_id);

		// The allocation site is already registered and we found its
		// slot. Return the slot index.

		if (site_ra == ra) {
			return site_id;
		}

		// We reached an empty slot, i.e., the allocation site isn't yet
		// registered. Try to register it. If somebody else managed to grab
		// this slot in the meantime, keep looping. Otherwise return the
		// slot index.

		if (site_ra == NULL && ck_pr_cas_ptr(g_site_ras + site_id, NULL, (void *)ra)) {
			ck_pr_inc_32(&g_n_site_ras);
			return site_id;
		}

		site_id = (site_id + 1) & (MAX_SITES - 1);
	}

	// More than MAX_SITES call sites.
	cf_crash(CF_ALLOC, "too many call sites");
	// Not reached.
	return 0;
}

static uint32_t
hook_new_site_info_id(void)
{
	uint64_t info_id = ck_pr_faa_64(&g_n_site_infos, 1);

	if (info_id < MAX_SITES * MAX_THREADS) {
		return (uint32_t)info_id;
	}

	if (info_id == MAX_SITES * MAX_THREADS) {
		cf_ticker_warning(CF_ALLOC, "site info pool exhausted");
	}

	return 0;
}

// Get the info ID of the site_info record for the given site ID and the current
// thread. In case the current thread doesn't yet have a site_info record for the
// given site ID, a new site_info record is allocated.

static uint32_t
hook_get_site_info_id(uint32_t site_id)
{
	uint32_t info_id = g_thread_site_infos[site_id];

	// This thread encountered this allocation site before. We already
	// have a site info record.

	if (info_id != 0) {
		return info_id;
	}

	// This is the first time that this thread encounters this allocation
	// site. We need to allocate a site_info record.

	if ((info_id = hook_new_site_info_id()) == 0) {
		return 0;
	}

	site_info *info = g_site_infos + info_id;

	info->site_id = site_id;
	info->thread_id = cf_thread_sys_tid();
	info->size_lo = 0;
	info->size_hi = 0;

	g_thread_site_infos[site_id] = info_id;
	return info_id;
}

// Account for an allocation by the current thread for the allocation site
// with the given address.

static void
hook_handle_alloc(const void *ra, void *p, void *p_indent, size_t sz)
{
	if (p == NULL) {
		return;
	}

	size_t jem_sz = jem_sallocx(p, 0);

	uint32_t site_id = hook_get_site_id(ra);
	uint32_t info_id = hook_get_site_info_id(site_id);

	if (info_id != 0) {
		site_info *info = g_site_infos + info_id;

		size_t size_lo = info->size_lo;
		info->size_lo += jem_sz;

		// Carry?

		if (info->size_lo < size_lo) {
			++info->size_hi;
		}
	}

	uint8_t *data = (uint8_t *)p + jem_sz - sizeof(uint32_t);
	uint32_t *data32 = (uint32_t *)data;

	uint8_t *mark = (uint8_t *)p_indent + sz;
	size_t delta = (size_t)(data - mark);

	// Keep 0xffff as a marker for double free detection.

	if (delta > 0xfffe) {
		delta = 0;
	}

	*data32 = ((site_id << 16) | (uint32_t)delta) * MULT + 1;

	for (uint32_t i = 0; i < 4 && i < delta; ++i) {
		mark[i] = data[i];
	}
}

// Account for a deallocation by the current thread for the allocation
// site with the given address.

static void
hook_handle_free(const void *ra, void *p, size_t jem_sz)
{
	uint8_t *data = (uint8_t *)p + jem_sz - sizeof(uint32_t);
	uint32_t *data32 = (uint32_t *)data;

	uint32_t val = (*data32 - 1) * MULT_INV;
	uint32_t site_id = val >> 16;
	uint32_t delta = val & 0xffff;

	if (site_id >= MAX_SITES) {
		cf_crash(CF_ALLOC, "corruption %zu@%p RA %p, invalid site ID", jem_sz, p, ra);
	}

	if (delta == 0xffff) {
		cf_crash(CF_ALLOC, "corruption %zu@%p RA 0x%lx, potential double free, possibly freed before with RA 0x%lx",
				jem_sz, p, cf_log_strip_aslr(ra),
				cf_log_strip_aslr(ck_pr_load_ptr(g_site_ras + site_id)));
	}

	if (delta > jem_sz - sizeof(uint32_t)) {
		cf_crash(CF_ALLOC, "corruption %zu@%p RA 0x%lx, invalid delta length, possibly allocated with RA 0x%lx",
				jem_sz, p, cf_log_strip_aslr(ra),
				cf_log_strip_aslr(ck_pr_load_ptr(g_site_ras + site_id)));
	}

	uint8_t *mark = data - delta;

	for (uint32_t i = 0; i < 4 && i < delta; ++i) {
		if (mark[i] != data[i]) {
			cf_crash(CF_ALLOC, "corruption %zu@%p RA 0x%lx, invalid mark, possibly allocated with RA 0x%lx",
					jem_sz, p, cf_log_strip_aslr(ra),
					cf_log_strip_aslr(ck_pr_load_ptr(g_site_ras + site_id)));
		}
	}

	uint32_t info_id = hook_get_site_info_id(site_id);

	if (info_id != 0) {
		site_info *info = g_site_infos + info_id;

		size_t size_lo = info->size_lo;
		info->size_lo -= jem_sz;

		// Borrow?

		if (info->size_lo > size_lo) {
			--info->size_hi;
		}
	}

	// Replace the allocation site with the deallocation site to facilitate
	// double-free debugging.

	site_id = hook_get_site_id(ra);

	// Also invalidate the delta length, so that we are more likely to detect
	// double frees.

	*data32 = ((site_id << 16) | 0xffff) * MULT + 1;

	for (uint32_t i = 0; i < 4 && i < delta; ++i) {
		mark[i] = data[i];
	}
}

static uint32_t
indent_hops(void *p)
{
	if (!g_rand.initialized) {
		g_rand.seed0 = (uint64_t)cf_thread_sys_tid();
		g_rand.seed1 = cf_getns();
		g_rand.initialized = true;
	}

	uint32_t n_hops;
	void **p_indent;

	// Indented pointer must not look like aligned allocation. See outdent().

	do {
		n_hops = 2 + (as_random_next_uint32(&g_rand) % ((MAX_INDENT / 8) - 1));
		p_indent = (void **)p + n_hops;
	}
	while (((uint64_t)p_indent & 0xfff) == 0);

	return n_hops;
}

static void *
indent(void *p)
{
	if (p == NULL) {
		return NULL;
	}

	uint32_t n_hops = indent_hops(p);
	uint64_t *p_indent = (uint64_t *)p + n_hops;

	p_indent[-1] = (uint64_t)p * MULT_64;
	*(uint64_t *)p = (uint64_t)p_indent * MULT_64;

	return (void *)p_indent;
}

static void *
reindent(void *p2, size_t sz, void *p, void *p_indent)
{
	if (p2 == NULL) {
		return NULL;
	}

	uint32_t n_hops = (uint32_t)(((uint8_t *)p_indent - (uint8_t *)p)) / 8;
	void **from = (void **)p2 + n_hops;

	uint32_t n_hops2 = indent_hops(p2);
	uint64_t *p2_indent = (uint64_t *)p2 + n_hops2;

	memmove(p2_indent, from, sz);

	p2_indent[-1] = (uint64_t)p2 * MULT_64;
	*(uint64_t *)p2 = (uint64_t)p2_indent * MULT_64;

	return (void *)p2_indent;
}

static void *
outdent(void *p_indent)
{
	// Aligned allocations aren't indented.

	if (((uint64_t)p_indent & 0xfff) == 0) {
		return p_indent;
	}

	uint64_t p = ((uint64_t *)p_indent)[-1] * MULT_INV_64;
	int64_t diff = (int64_t)p_indent - (int64_t)p;

	if (diff < 16 || diff > MAX_INDENT || diff % 8 != 0) {
		cf_crash(CF_ALLOC, "bad free of %p via %p", (void *)p, p_indent);
	}

	uint64_t p_expect = *(uint64_t *)p * MULT_INV_64;

	if ((uint64_t)p_indent != p_expect) {
		cf_crash(CF_ALLOC, "bad free of %p via %p (vs. %p)", (void *)p,
				p_indent, (void *)p_expect);
	}

	return (void *)p;
}

static void
valgrind_check(void)
{
	// Make sure that we actually call into JEMalloc when invoking malloc().
	//
	// By default, Valgrind redirects the standard allocation API functions,
	// i.e., malloc(), calloc(), etc., to glibc.
	//
	// The problem with this is that Valgrind only redirects the standard API
	// functions. It does not know about, and thus doesn't redirect, our
	// non-standard functions, e.g., cf_alloc_malloc_arena().
	//
	// As we use both, standard and non-standard functions, to allocate memory,
	// we would end up with an inconsistent mix of allocations, some allocated
	// by JEMalloc and some by glibc's allocator.
	//
	// Sooner or later, we will thus end up passing a memory block allocated by
	// JEMalloc to free(), which Valgrind has redirected to glibc's allocator.

	uint32_t tries;

	void *p1[2];
	void *p2[2];

	for (tries = 0; tries < 2; ++tries) {
		p1[tries] = malloc(1); // known API function, possibly redirected
		p2[tries] = cf_alloc_try_malloc(1); // our own, never redirected

		// If both of the above allocations are handled by JEMalloc, then their
		// base addresses will be identical (cache enabled), contiguous (cache
		// disabled), or unrelated (cache disabled, different runs). Trying
		// twice prevents the latter.
		//
		// If the first allocation is handled by glibc, then the base addresses
		// will always be unrelated.

		ptrdiff_t diff = (uint8_t *)p2[tries] - (uint8_t *)p1[tries];

		if (diff > -1024 && diff < 1024) {
			break;
		}
	}

	if (tries == 2) {
		cf_crash_nostack(CF_ALLOC, "Valgrind redirected malloc() to glibc; please run Valgrind with --soname-synonyms=somalloc=nouserintercepts");
	}

	for (uint32_t i = 0; i < tries; ++i) {
		free(p1[tries]);
		cf_free(p2[tries]);
	}
}

void
cf_alloc_init(void)
{
	valgrind_check();

	// Turn off libstdc++'s memory caching, as it just duplicates JEMalloc's.

	if (setenv("GLIBCXX_FORCE_NEW", "1", 1) < 0) {
		cf_crash(CF_ALLOC, "setenv() failed: %d (%s)", errno, cf_strerror(errno));
	}

	// Double-check that hook_get_arena() works, as it depends on JEMalloc's
	// internal data structures.

	int32_t err = jem_mallctl("thread.tcache.flush", NULL, NULL, NULL, 0);

	if (err != 0) {
		cf_crash(CF_ALLOC, "error while flushing thread cache: %d (%s)", err, cf_strerror(err));
	}

	for (size_t sz = 1; sz <= 16 * 1024 * 1024; sz *= 2) {
		void *p = malloc(sz);
		int32_t arena = hook_get_arena(p);

		if (arena != N_ARENAS) {
			cf_crash(CF_ALLOC, "arena mismatch: %d vs. %d", arena, N_ARENAS / 2);
		}

		free(p);
	}
}

void
cf_alloc_set_debug(cf_alloc_debug debug_allocations, bool indent_allocations)
{
	g_debug = debug_allocations;
	g_indent = indent_allocations;

	g_alloc_started = true;
}

int32_t
cf_alloc_create_arena(void)
{
	int32_t arena;
	size_t arena_len = sizeof(arena);

	int32_t err = jem_mallctl("arenas.extend", &arena, &arena_len, NULL, 0);

	if (err != 0) {
		cf_crash(CF_ALLOC, "failed to create new arena: %d (%s)", err, cf_strerror(err));
	}

	cf_debug(CF_ALLOC, "created new arena %d", arena);
	return arena;
}

void
cf_alloc_heap_stats(size_t *allocated_kbytes, size_t *active_kbytes, size_t *mapped_kbytes,
		double *efficiency_pct, uint32_t *site_count)
{
	uint64_t epoch = 1;
	size_t len = sizeof(epoch);

	int32_t err = jem_mallctl("epoch", &epoch, &len, &epoch, len);

	if (err != 0) {
		cf_crash(CF_ALLOC, "failed to retrieve epoch: %d (%s)", err, cf_strerror(err));
	}

	size_t allocated;
	len = sizeof(allocated);

	err = jem_mallctl("stats.allocated", &allocated, &len, NULL, 0);

	if (err != 0) {
		cf_crash(CF_ALLOC, "failed to retrieve stats.allocated: %d (%s)", err, cf_strerror(err));
	}

	size_t active;
	len = sizeof(active);

	err = jem_mallctl("stats.active", &active, &len, NULL, 0);

	if (err != 0) {
		cf_crash(CF_ALLOC, "failed to retrieve stats.active: %d (%s)", err, cf_strerror(err));
	}

	size_t mapped;
	len = sizeof(mapped);

	err = jem_mallctl("stats.mapped", &mapped, &len, NULL, 0);

	if (err != 0) {
		cf_crash(CF_ALLOC, "failed to retrieve stats.mapped: %d (%s)", err, cf_strerror(err));
	}

	if (allocated_kbytes) {
		*allocated_kbytes = allocated / 1024;
	}

	if (active_kbytes) {
		*active_kbytes = active / 1024;
	}

	if (mapped_kbytes) {
		*mapped_kbytes = mapped / 1024;
	}

	if (efficiency_pct) {
		*efficiency_pct = mapped != 0 ?
				(double)allocated * 100.0 / (double)mapped : 0.0;
	}

	if (site_count) {
		*site_count = ck_pr_load_32(&g_n_site_ras);
	}
}

static void
line_to_log(void *data, const char *line)
{
	(void)data;

	char buff[1000];
	size_t i;

	for (i = 0; i < sizeof(buff) - 1 && line[i] != 0 && line[i] != '\n'; ++i) {
		buff[i] = line[i];
	}

	buff[i] = 0;
	cf_info(CF_ALLOC, "%s", buff);
}

static void
line_to_file(void *data, const char *line)
{
	fprintf((FILE *)data, "%s", line);
}

static void
time_to_file(FILE *fh)
{
	time_t now = time(NULL);

	if (now == (time_t)-1) {
		cf_crash(CF_ALLOC, "time() failed: %d (%s)", errno, cf_strerror(errno));
	}

	struct tm gmt;

	if (gmtime_r(&now, &gmt) == NULL) {
		cf_crash(CF_ALLOC, "gmtime_r() failed");
	}

	char text[250];

	if (strftime(text, sizeof(text), "%b %d %Y %T %Z", &gmt) == 0) {
		cf_crash(CF_ALLOC, "strftime() failed");
	}

	fprintf(fh, "---------- %s ----------\n", text);
}

void
cf_alloc_log_stats(const char *file, const char *opts)
{
	if (file == NULL) {
		jem_malloc_stats_print(line_to_log, NULL, opts);
		return;
	}

	FILE *fh = fopen(file, "a");

	if (fh == NULL) {
		cf_warning(CF_ALLOC, "failed to open allocation stats file %s: %d (%s)",
				file, errno, cf_strerror(errno));
		return;
	}

	time_to_file(fh);
	jem_malloc_stats_print(line_to_file, fh, opts);
	fclose(fh);
}

void
cf_alloc_log_site_infos(const char *file)
{
	FILE *fh = fopen(file, "a");

	if (fh == NULL) {
		cf_warning(CF_ALLOC, "failed to open site info file %s: %d (%s)",
				file, errno, cf_strerror(errno));
		return;
	}

	time_to_file(fh);
	uint64_t n_site_infos = ck_pr_load_64(&g_n_site_infos);

	if (n_site_infos > MAX_SITES * MAX_THREADS) {
		n_site_infos = MAX_SITES * MAX_THREADS;
	}

	for (uint64_t i = 1; i < n_site_infos; ++i) {
		site_info *info = g_site_infos + i;
		const void *ra = ck_pr_load_ptr(g_site_ras + info->site_id);

		fprintf(fh, "0x%016" PRIx64 " %9d 0x%016zx 0x%016zx\n",
				cf_log_strip_aslr(ra), info->thread_id, info->size_hi,
				info->size_lo);
	}

	fclose(fh);
}

static bool
is_transient(int32_t arena)
{
	// Note that this also considers -1 (i.e., the default thread arena)
	// to be transient, in addition to arenas 0 .. (N_ARENAS - 1).

	return arena < N_ARENAS;
}

static bool
want_debug(int32_t arena)
{
	// No debugging during startup and for startup arena.

	if (!g_alloc_started || arena == N_ARENAS) {
		return false;
	}

	switch (g_debug) {
	case CF_ALLOC_DEBUG_NONE:
		return false;

	case CF_ALLOC_DEBUG_TRANSIENT:
		return is_transient(arena);

	case CF_ALLOC_DEBUG_PERSISTENT:
		return !is_transient(arena);

	case CF_ALLOC_DEBUG_ALL:
		return true;
	}

	// Not reached.
	return false;
}

static int32_t
calc_free_flags(int32_t arena)
{
	cf_assert(g_alloc_started || arena == N_ARENAS, CF_ALLOC,
			"bad arena %d during startup", arena);

	// Bypass the thread-local cache for allocations in the startup arena.

	if (arena == g_startup_arena) {
		return MALLOCX_TCACHE_NONE;
	}

	// If it's a transient allocation, then simply use the default
	// thread-local cache. No flags needed. Same, if we don't debug
	// at all; then we can save ourselves the second cache.

	if (is_transient(arena) || g_debug == CF_ALLOC_DEBUG_NONE) {
		return 0;
	}

	// If it's a persistent allocation, then use the second per-thread
	// cache. Add it to the flags. See calc_alloc_flags() for more on
	// this second cache.

	return MALLOCX_TCACHE(g_ns_tcache);
}

static void
do_free(void *p_indent, const void *ra)
{
	if (p_indent == NULL) {
		return;
	}

	int32_t arena = hook_get_arena(p_indent);
	int32_t flags = calc_free_flags(arena);

	if (!want_debug(arena)) {
		jem_dallocx(p_indent, flags); // not indented
		return;
	}

	void *p = g_indent ? outdent(p_indent) : p_indent;
	size_t jem_sz = jem_sallocx(p, 0);

	hook_handle_free(ra, p, jem_sz);
	jem_sdallocx(p, jem_sz, flags);
}

void
__attribute__ ((noinline))
free(void *p_indent)
{
	do_free(p_indent, __builtin_return_address(0));
}

static int32_t
calc_alloc_flags(int32_t flags, int32_t arena)
{
	cf_assert(g_alloc_started || arena < 0, CF_ALLOC,
			"bad arena %d during startup", arena);

	// During startup, allocate from the startup arena and bypass the
	// thread-local cache.

	if (!g_alloc_started) {
		// Create startup arena, if necessary.
		if (g_startup_arena < 0) {
			size_t len = sizeof(g_startup_arena);
			int32_t err = jem_mallctl("arenas.extend", &g_startup_arena, &len,
					NULL, 0);

			if (err != 0) {
				cf_crash(CF_ALLOC, "failed to create startup arena: %d (%s)",
						err, cf_strerror(err));
			}

			// Expect arena 149.
			cf_assert(g_startup_arena == N_ARENAS, CF_ALLOC,
					"bad startup arena %d", g_startup_arena);
		}

		// Set startup arena, bypass thread-local cache.
		flags |= MALLOCX_ARENA(g_startup_arena) | MALLOCX_TCACHE_NONE;

		return flags;
	}

	// Default arena and default thread-local cache. No additional flags
	// needed.

	if (arena < 0) {
		return flags;
	}

	// We're allocating from a specific arena. Add it to the flags.

	flags |= MALLOCX_ARENA(arena);

	// If it's an arena for transient allocations, then we use the default
	// thread-local cache. No additional flags needed. Same, if we don't
	// debug at all; then we can save ourselves the second cache.

	if (is_transient(arena) || g_debug == CF_ALLOC_DEBUG_NONE) {
		return flags;
	}

	// We have a second per-thread cache for persistent allocations. In this
	// way we never mix persistent allocations and transient allocations in
	// the same cache. We need to keep them apart, because debugging may be
	// enabled for one, but not the other.

	// Create the second per-thread cache, if we haven't already done so.

	if (g_ns_tcache < 0) {
		size_t len = sizeof(g_ns_tcache);
		int32_t err = jem_mallctl("tcache.create", &g_ns_tcache, &len, NULL, 0);

		if (err != 0) {
			cf_crash(CF_ALLOC, "failed to create new cache: %d (%s)", err, cf_strerror(err));
		}
	}

	// Add the second (non-default) per-thread cache to the flags.

	flags |= MALLOCX_TCACHE(g_ns_tcache);
	return flags;
}

static void *
do_mallocx(size_t sz, int32_t arena, const void *ra)
{
	int32_t flags = calc_alloc_flags(0, arena);

	if (!want_debug(arena)) {
		return jem_mallocx(sz == 0 ? 1 : sz, flags);
	}

	size_t ext_sz = sz + sizeof(uint32_t);

	if (g_indent) {
		ext_sz += MAX_INDENT;
	}

	void *p = jem_mallocx(ext_sz, flags);
	void *p_indent = g_indent ? indent(p) : p;

	hook_handle_alloc(ra, p, p_indent, sz);

	return p_indent;
}

void *
cf_alloc_try_malloc(size_t sz)
{
	// Allowed to return NULL.
	return do_mallocx(sz, -1, __builtin_return_address(0));
}

void *
cf_alloc_malloc_arena(size_t sz, int32_t arena)
{
	cf_assert(g_alloc_started, CF_ALLOC, "arena allocation during startup");

	void *p_indent = do_mallocx(sz, arena, __builtin_return_address(0));

	cf_assert(p_indent != NULL, CF_ALLOC, "malloc_ns failed sz %zu arena %d",
			sz, arena);

	return p_indent;
}

void *
__attribute__ ((noinline))
malloc(size_t sz)
{
	void *p_indent = do_mallocx(sz, -1, __builtin_return_address(0));

	cf_assert(p_indent != NULL, CF_ALLOC, "malloc failed sz %zu", sz);

	return p_indent;
}

static void *
do_callocx(size_t n, size_t sz, int32_t arena, const void *ra)
{
	int32_t flags = calc_alloc_flags(MALLOCX_ZERO, arena);
	size_t tot_sz = n * sz;

	if (!want_debug(arena)) {
		return jem_mallocx(tot_sz == 0 ? 1 : tot_sz, flags);
	}

	size_t ext_sz = tot_sz + sizeof(uint32_t);

	if (g_indent) {
		ext_sz += MAX_INDENT;
	}

	void *p = jem_mallocx(ext_sz, flags);
	void *p_indent = g_indent ? indent(p) : p;

	hook_handle_alloc(ra, p, p_indent, tot_sz);

	return p_indent;
}

void *
cf_alloc_calloc_arena(size_t n, size_t sz, int32_t arena)
{
	cf_assert(g_alloc_started, CF_ALLOC, "arena allocation during startup");

	void *p_indent = do_callocx(n, sz, arena, __builtin_return_address(0));

	cf_assert(p_indent != NULL, CF_ALLOC,
			"calloc_ns failed n %zu sz %zu arena %d", n, sz, arena);

	return p_indent;
}

void *
calloc(size_t n, size_t sz)
{
	void *p_indent = do_callocx(n, sz, -1, __builtin_return_address(0));

	cf_assert(p_indent != NULL, CF_ALLOC, "calloc failed n %zu sz %zu", n, sz);

	return p_indent;
}

static void *
do_rallocx(void *p_indent, size_t sz, int32_t arena, const void *ra)
{
	if (p_indent == NULL) {
		return do_mallocx(sz, arena, ra);
	}

	int32_t arena_p = hook_get_arena(p_indent);

	bool debug_p = want_debug(arena_p);
	bool debug = want_debug(arena);

	// Allow debug change for startup arena - handled below.

	if (debug != debug_p && arena_p != N_ARENAS) {
		cf_crash(CF_ALLOC, "debug change - p_indent %p arena_p %d arena %d",
				p_indent, arena_p, arena);
	}

	if (sz == 0) {
		do_free(p_indent, ra);
		return NULL;
	}

	int32_t flags = calc_alloc_flags(0, arena);

	// Going from startup or non-debug arena to non-debug arena.

	if (!debug) {
		return jem_rallocx(p_indent, sz, flags); // not indented
	}

	// Going from startup arena to debug arena.

	if (arena_p == N_ARENAS) {
		void *p = p_indent; // not indented
		void *p_move = do_mallocx(sz, arena, ra);

		size_t sz_move = jem_sallocx(p, 0);

		if (sz < sz_move) {
			sz_move = sz;
		}

		memcpy(p_move, p, sz_move);
		cf_free(p);

		return p_move;
	}

	// Going from debug arena to debug arena.

	void *p = g_indent ? outdent(p_indent) : p_indent;
	size_t jem_sz = jem_sallocx(p, 0);

	hook_handle_free(ra, p, jem_sz);

	size_t ext_sz = sz + sizeof(uint32_t);

	if (g_indent) {
		ext_sz += MAX_INDENT;
	}

	void *p2 = jem_rallocx(p, ext_sz, flags);
	void *p2_indent = g_indent ? reindent(p2, sz, p, p_indent) : p2;

	hook_handle_alloc(ra, p2, p2_indent, sz);

	return p2_indent;
}

void *
cf_alloc_realloc_arena(void *p_indent, size_t sz, int32_t arena)
{
	cf_assert(g_alloc_started, CF_ALLOC, "arena allocation during startup");

	void *p2_indent = do_rallocx(p_indent, sz, arena,
			__builtin_return_address(0));

	cf_assert(p2_indent != NULL || sz == 0, CF_ALLOC,
			"realloc_ns failed sz %zu arena %d", sz, arena);

	return p2_indent;
}

void *
realloc(void *p_indent, size_t sz)
{
	void *p2_indent = do_rallocx(p_indent, sz, -1, __builtin_return_address(0));

	cf_assert(p2_indent != NULL || sz == 0, CF_ALLOC, "realloc failed sz %zu",
			sz);

	return p2_indent;
}

static char *
do_strdup(const char *s, size_t n, const void *ra)
{
	int32_t flags = calc_alloc_flags(0, -1);

	size_t sz = n + 1;
	size_t ext_sz = sz;

	if (want_debug(-1)) {
		ext_sz += sizeof(uint32_t);

		if (g_indent) {
			ext_sz += MAX_INDENT;
		}
	}

	char *s2 = jem_mallocx(ext_sz, flags);
	char *s2_indent = s2;

	if (want_debug(-1)) {
		if (g_indent) {
			s2_indent = indent(s2);
		}

		hook_handle_alloc(ra, s2, s2_indent, sz);
	}

	memcpy(s2_indent, s, n);
	s2_indent[n] = 0;

	return s2_indent;
}

char *
strdup(const char *s)
{
	return do_strdup(s, strlen(s), __builtin_return_address(0));
}

char *
strndup(const char *s, size_t n)
{
	size_t n2 = 0;

	while (n2 < n && s[n2] != 0) {
		++n2;
	}

	return do_strdup(s, n2, __builtin_return_address(0));
}

static int32_t
do_asprintf(char **res, const char *form, va_list va, const void *ra)
{
	char buff[25000];
	int32_t n = vsnprintf(buff, sizeof(buff), form, va);

	if ((size_t)n >= sizeof(buff)) {
		cf_crash(CF_ALLOC, "asprintf overflow len %d", n);
	}

	*res = do_strdup(buff, (size_t)n, ra);
	return n;
}

int32_t
asprintf(char **res, const char *form, ...)
{
	va_list va;
	va_start(va, form);

	int32_t n = do_asprintf(res, form, va, __builtin_return_address(0));

	va_end(va);
	return n;
}

int32_t
__asprintf_chk(char **res, int32_t flags, const char *form, ...)
{
	(void)flags;

	va_list va;
	va_start(va, form);

	int32_t n = do_asprintf(res, form, va, __builtin_return_address(0));

	va_end(va);
	return n;
}

int32_t
posix_memalign(void **p, size_t align, size_t sz)
{
	cf_assert(g_alloc_started, CF_ALLOC, "aligned allocation during startup");
	cf_assert((align & (align - 1)) == 0, CF_ALLOC, "bad alignment");

	if (!want_debug(-1)) {
		return jem_posix_memalign(p, align, sz == 0 ? 1 : sz);
	}

	if (g_indent) {
		align = (align + 0xffful) & ~0xffful;
	}

	size_t ext_sz = sz + sizeof(uint32_t);
	int32_t err = jem_posix_memalign(p, align, ext_sz);

	if (err != 0) {
		return err;
	}

	hook_handle_alloc(__builtin_return_address(0), *p, *p, sz);
	return 0;
}

void *
aligned_alloc(size_t align, size_t sz)
{
	cf_assert(g_alloc_started, CF_ALLOC, "aligned allocation during startup");
	cf_assert((align & (align - 1)) == 0, CF_ALLOC, "bad alignment");

	if (!want_debug(-1)) {
		return jem_aligned_alloc(align, sz == 0 ? 1 : sz);
	}

	if (g_indent) {
		align = (align + 0xffful) & ~0xffful;
	}

	size_t ext_sz = sz + sizeof(uint32_t);

	void *p = jem_aligned_alloc(align, ext_sz);
	hook_handle_alloc(__builtin_return_address(0), p, p, sz);

	return p;
}

static void *
do_valloc(size_t sz)
{
	if (!want_debug(-1)) {
		return jem_aligned_alloc(PAGE_SZ, sz == 0 ? 1 : sz);
	}

	size_t ext_sz = sz + sizeof(uint32_t);

	void *p = jem_aligned_alloc(PAGE_SZ, ext_sz);
	hook_handle_alloc(__builtin_return_address(0), p, p, sz);

	return p;
}

void *
valloc(size_t sz)
{
	cf_assert(g_alloc_started, CF_ALLOC, "aligned allocation during startup");

	void *p = do_valloc(sz);
	cf_assert(p, CF_ALLOC, "valloc failed sz %zu", sz);
	return p;
}

void *
memalign(size_t align, size_t sz)
{
	cf_assert(g_alloc_started, CF_ALLOC, "aligned allocation during startup");
	cf_assert((align & (align - 1)) == 0, CF_ALLOC, "bad alignment");

	if (!want_debug(-1)) {
		return jem_aligned_alloc(align, sz == 0 ? 1 : sz);
	}

	if (g_indent) {
		align = (align + 0xffful) & ~0xffful;
	}

	size_t ext_sz = sz + sizeof(uint32_t);

	void *p = jem_aligned_alloc(align, ext_sz);
	hook_handle_alloc(__builtin_return_address(0), p, p, sz);

	return p;
}

void *
pvalloc(size_t sz)
{
	(void)sz;
	cf_crash(CF_ALLOC, "obsolete pvalloc() called");
	// Not reached.
	return NULL;
}

void *
cf_rc_alloc(size_t sz)
{
	int32_t flags = calc_alloc_flags(0, -1);

	size_t tot_sz = sizeof(cf_rc_header) + sz;
	size_t ext_sz = tot_sz;

	if (want_debug(-1)) {
		ext_sz += sizeof(uint32_t);

		if (g_indent) {
			ext_sz += MAX_INDENT;
		}
	}

	void *p = jem_mallocx(ext_sz, flags);
	void *p_indent = p;

	if (want_debug(-1)) {
		if (g_indent) {
			p_indent = indent(p);
		}

		hook_handle_alloc(__builtin_return_address(0), p, p_indent, tot_sz);
	}

	cf_rc_header *head = p_indent;

	head->rc = 1;
	head->sz = (uint32_t)sz;

	return head + 1; // body
}

static void
do_rc_free(void *body, void *ra)
{
	if (body == NULL) {
		cf_crash(CF_ALLOC, "trying to cf_rc_free() null pointer");
	}

	cf_rc_header *head = (cf_rc_header *)body - 1;

	int32_t arena = hook_get_arena(head);
	int32_t flags = calc_free_flags(arena);

	if (!want_debug(arena)) {
		jem_dallocx(head, flags); // not indented
		return;
	}

	void *p = g_indent ? outdent(head) : head;
	size_t jem_sz = jem_sallocx(p, 0);

	hook_handle_free(ra, p, jem_sz);
	jem_sdallocx(p, jem_sz, flags);
}

void
cf_rc_free(void *body)
{
	do_rc_free(body, __builtin_return_address(0));
}

int32_t
cf_rc_reserve(void *body)
{
	cf_rc_header *head = (cf_rc_header *)body - 1;
	return cf_atomic32_incr(&head->rc);
}

int32_t
cf_rc_release(void *body)
{
	cf_rc_header *head = (cf_rc_header *)body - 1;
	int32_t rc = cf_atomic32_decr(&head->rc);

	cf_assert(rc >= 0, CF_ALLOC, "reference count underflow: %d (0x%x)", rc,
			rc);

	return rc;
}

int32_t
cf_rc_releaseandfree(void *body)
{
	cf_rc_header *head = (cf_rc_header *)body - 1;
	int32_t rc = cf_atomic32_decr(&head->rc);

	cf_assert(rc >= 0, CF_ALLOC, "reference count underflow: %d (0x%x)", rc,
			rc);

	if (rc > 0) {
		return rc;
	}

	do_rc_free(body, __builtin_return_address(0));
	return 0;
}

int32_t
cf_rc_count(const void *body)
{
	const cf_rc_header *head = (const cf_rc_header *)body - 1;
	return (int32_t)head->rc;
}
